
// Project: webCDwriter - Network CD Writing
// File: Process.cpp
// Author: Jrg Haeger, 02.11.1999

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>

#include "Support.h"
#include "Log.h"
#include "Process.h"

Process::Process(const char *fileName, const char *protoName)
{
	argsNum = envsNum = 0;
	args[argsNum++] = strDup(fileName);
	args[argsNum] = NULL;
	envs[envsNum] = NULL;
	pid = -1;
	result = -1;
	outPipe[0] = outPipe[1] = 0;
	inPipe[0] = inPipe[1] = 0;
	errPipe[0] = errPipe[1] = 0;
	inStream = NULL;
	if (protoName == NULL)
		protoStream = NULL;
	else
		protoStream = fopen(protoName, "w");
	linePos = 0;
	lineStr[0] = 0;
}

Process::~Process()
{
	getResult();
	log.debug("Process", "~Process() %s %d result = %d\n",
		args[0], pid, result);

	if (protoStream != NULL)
		fclose(protoStream);
	if (inStream != NULL)
		fclose(inStream);
	for (int i = 0; i < envsNum; i++)
		delete envs[i];
	for (int i = 0; i < argsNum; i++)
		delete args[i];
	closePipes();
}

void Process::addArg(const char *argStr)
{
	if (argsNum == 256 - 1)
		throw "Process::addArg: to many args";

	args[argsNum++] = strDup(argStr);
	args[argsNum] = NULL;
}

void Process::closePipes()
{
	for (int i = 0; i < 2; i++)
	if (outPipe[i] > 0)
	{
		close(outPipe[i]);
		outPipe[i] = 0;
	}

	for (int i = 0; i < 2; i++)
	if (inPipe[i] > 0)
	{
		close(inPipe[i]);
		inPipe[i] = 0;
	}

	for (int i = 0; i < 2; i++)
	if (errPipe[i] > 0)
	{
		close(errPipe[i]);
		errPipe[i] = 0;
	}
}

void Process::closePipeToChild()
{
	close(outPipe[STDOUT_FILENO]);
	outPipe[STDOUT_FILENO] = 0;
}

int Process::getResult()
{
	if (result >= 0 || pid <= 0)
		return result;

	if (inStream != NULL)
		fclose(inStream);
	inStream = NULL;

	int res, status;
//	TEMP_FAILURE_RETRY(waitpid(pid, &status, 0));
	while ((res = waitpid(pid, &status, 0)) == -1 && errno == EINTR);

	if (res == -1)
		result = 256;
	else
	if (WIFEXITED(status))
		result = WEXITSTATUS(status);
	else
		result = 257;
	return result;
}

int Process::read(char *buf, int count)
{
	fcntl(inPipe[STDIN_FILENO], F_SETFL, O_NONBLOCK);
	return ::read(inPipe[STDIN_FILENO], buf, count);
}

const char *Process::readLine()
{
	while (1)
	{
		int ch = fgetc(inStream);
		if (ch == 8)
		{
			if (linePos > 0)
				linePos--;
		}
		else
		if (ch == '\r')
			linePos = 0;
		else
		if (ch == '\n')
		{
			linePos = 0;
			lineStr[0] = 0;
		}
		else
		{
			lineStr[linePos++] = ch;
			break;
		}
	}

	while (linePos < sizeof lineStr - 2)
	{
		int ch = fgetc(inStream);
		if (ch == EOF)
			return NULL;
		if (ch == 8 || ch == '\r' || ch == '\n')
		{
			ungetc(ch, inStream);
			break;
		}
		lineStr[linePos++] = ch;
	}
	lineStr[linePos] = '\n';;
	lineStr[linePos+1] = 0;
	if (protoStream != NULL)
		fputs(lineStr, protoStream);
	return lineStr;
}

int Process::run()
{
	start();
	while (1)
	{
		const char *lineStr = readLine();
		if (lineStr == NULL)
			break;
	}
	return getResult();
}

void Process::select()
{
	fd_set fdSet;
	FD_ZERO(&fdSet);
	FD_SET(inPipe[STDIN_FILENO], &fdSet);

	fd_set fdWriteSet;
	FD_ZERO(&fdWriteSet);
	FD_SET(outPipe[STDOUT_FILENO], &fdWriteSet);

	::select(FD_SETSIZE, &fdSet, &fdWriteSet, NULL, NULL);
}

void Process::show()
{
	log.putv("Process: <%s>\n", args[0]);
	for (int i = 1; i < argsNum; i++)
		log.putv("%8d: <%s>\n", i, args[i]);
}

/**
	@see glibc-2.1.1/sysdeps/posix/pipestream.c
*/
void Process::start(const char *workDirName, int splitOut)
{
	if (pipe(outPipe) != 0)
		throw "outPipe failed";
	if (pipe(inPipe) != 0)
		throw "inPipe failed";
	if (splitOut)
		if (pipe(errPipe) != 0)
			throw "errPipe failed";

	pid = fork();
	if (pid == (pid_t)-1)
		throw "fork failed";

	if (pid == (pid_t)0)
	{
		// We are the child side.
		dup2(outPipe[STDIN_FILENO], STDIN_FILENO);
		dup2(inPipe[STDOUT_FILENO], STDOUT_FILENO);
		if (splitOut)
			dup2(errPipe[STDOUT_FILENO], STDERR_FILENO);
		else
			dup2(inPipe[STDOUT_FILENO], STDERR_FILENO);

		// Close the pipe descriptors.
		closePipes();

		if (workDirName != NULL)
			if (chdir(workDirName) != 0)
				exit(129);

//		execve(args[0], args, envs);
		execvp(args[0], args);
		exit(128);
	}

	log.debug("Process", "start() %s %d\n", args[0], pid);
	close(outPipe[STDIN_FILENO]);
	outPipe[STDIN_FILENO] = 0;
	close(inPipe[STDOUT_FILENO]);
	inPipe[STDOUT_FILENO] = 0;
	if (errPipe[STDOUT_FILENO] > 0)
	{
		close(errPipe[STDOUT_FILENO]);
		errPipe[STDOUT_FILENO] = 0;
	}

	if (splitOut)
	{
		inStream = fdopen(errPipe[STDIN_FILENO], "r");
		errPipe[STDIN_FILENO] = 0;
	}
	else
	{
		inStream = fdopen(inPipe[STDIN_FILENO], "r");
		inPipe[STDIN_FILENO] = 0;
	}
	if (inStream == NULL)
		throw "open inStream failed";
}

void Process::terminate()
{
	if (pid > 0)
	{
		kill(pid, SIGTERM);
		getResult();
	}
}

const char *Process::toString(char *str)
{
	sprintf(str, "Process: <%s", args[0]);
	for (int i = 1; i < argsNum; i++)
	{
		strcat(str, " ");
		strcat(str, args[i]);
	}
	strcat(str, ">");
	return str;
}

int Process::write(const char *buf, int count)
{
	fcntl(outPipe[STDOUT_FILENO], F_SETFL, O_NONBLOCK);
	return ::write(outPipe[STDOUT_FILENO], buf, count);
}

