
// Project: webCDwriter - Network CD Writing
// File: Session.cpp
// Author: Jrg Haeger, 05.06.1999

#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <utime.h>

#include "Config.h"
#include "Support.h"
#include "Log.h"
#include "MP3toWAV.h"
#include "Queue.h"
#include "Server.h"
#include "Session.h"
#include "Store.h"

const char
	CDWserverInfo[] = "CDWserver.info";

Session::Session(Server *server, const char *sessionID)
{
	this->server = server;
	ID = strDup(sessionID);
	filesNum = tracksNum = sizeInKBytes = 0;
	isOpen = 0;

	dir = store->registerSession(this);
	infoFile = new File(*dir, CDWserverInfo);
	currentTrack = NULL;
	currentTrackNo = 0;
	mp3towav = NULL;
}

Session::~Session()
{
	if (mp3towav != NULL)
		closeMP3Track();

	File file(dir, CDWserverInfo);
	FILE *outStream = fopen(file.getName(), "w");
	if (outStream != NULL)
	{
		fprintf(outStream, "sizeInKBytes=%d\n", sizeInKBytes);
		fprintf(outStream, "files=%d\n", filesNum);
		fprintf(outStream, "tracks=%d\n", tracksNum);
		fclose(outStream);
	}

	store->unregister(this);
	delete[] ID;
}

void Session::closeMP3Track()
{
	log.putv("Session::closeMP3Track()");
	mp3towav->end();
	FILE *outStream = fopen(currentTrack->getName(), "r+");
	if (outStream != NULL)
	{
		fseek(outStream, 0, SEEK_END);
		try {
			while (1)
			{
				const char *buf;
				int n = mp3towav->get(&buf);
				if (n == 0)
					break;
				copy(buf, outStream, n);
			}
		} catch (const char *str)
		{
		}
		mp3towav->updateHeader(outStream);
		fclose(outStream);
	}

	delete mp3towav;
	mp3towav = NULL;
}

void Session::copy(FILE *inStream, FILE *outStream, int length)
{
	try {
		store->announceKBytes(sizeToKBytes(length));
		copy2(inStream, outStream, length);
	}
	catch (const char *msg)
	{
		store->cancelKBytes(sizeToKBytes(length));
		skipBytes(inStream, length);
		throw msg;
	}
}

void Session::copy(const char *buf, FILE *outStream, int length)
{
	int KBytes = sizeToKBytes(length);
	if (sizeInKBytes + KBytes > 1024 * config.getMaxMBytesPerSession())
		throw "session to big";

	store->announceKBytes(sizeToKBytes(length));

	int w = 0;
	while (w < length)
	{
		w += fwrite(&buf[w], 1, length - w, outStream);
		store->announceKBytes(0);
	}
	sizeInKBytes += length / 1024;
	store->registerKBytes(length / 1024);
}

void Session::copy2(FILE *inStream, FILE *outStream, int &length)
{
	int KBytes = sizeToKBytes(length);
	if (sizeInKBytes + KBytes > 1024 * config.getMaxMBytesPerSession())
		throw "session to big";

	while (length > 0)
	{
		char buf[16 * 1024];
		int n = fread(buf, 1,
			min(length, sizeof buf), inStream);
		length -= n;
		if (feof(inStream))
			throw "unexpected end of inStream";
		int w = 0;
		while (w < n)
		{
			w += fwrite(&buf[w], 1, n - w, outStream);
			store->announceKBytes(0);
		}
		sizeInKBytes += n / 1024;
		store->registerKBytes(n / 1024);
	}
}

void Session::countFile(const char *pathName)
{
	struct stat statBuf;
	if (lstat(pathName, &statBuf) != 0)
		return;

	if (!S_ISDIR(statBuf.st_mode))
	{
		if (!S_ISREG(statBuf.st_mode))
			return;
		filesNum++;
		sizeInKBytes += sizeToKBytes(statBuf.st_size);
		return;
	}

	DIR *dir = opendir(pathName);
	if (dir == NULL)
		return;

	struct dirent *dirEntry;
	while ((dirEntry = readdir(dir)) != NULL)
	{
		if (strcmp(dirEntry->d_name, ".") == 0
			|| strcmp(dirEntry->d_name, "..") == 0)
			continue;

		char tmpPathName[strlen(pathName) + 1
			+ strlen(dirEntry->d_name) + 1];
		sprintf(tmpPathName, "%s/%s",
			pathName, dirEntry->d_name);
		countFile(tmpPathName);
	}

	closedir(dir);
}

void Session::create()
{
	if (dir->exists())
		store->removeDir(dir->getName());

	infoFile->mkdirs();
	FILE *outStream = fopen(infoFile->getName(), "w");
	if (outStream == NULL)
		throw "internal error:"
			" cannot create infofile";
	fclose(outStream);

	isOpen = 1;
}

const File *Session::getFirstTrack()
{
	if (mp3towav != NULL)
		closeMP3Track();

	currentTrackNo = 0;
	return getNextTrack();
}

int Session::getSessionSize(File *dir)
{
	int sizeInKBytes = 0;
	File file(dir, CDWserverInfo);
	FILE *inStream = fopen(file.getName(), "r");
	if (inStream == NULL)
		throw "internal error: can't read sessionInfo";

	char lineStr[1024];
	while (fgets(lineStr, sizeof lineStr, inStream) != NULL)
	{
		char str1[1024], str2[1024];
		Config::parseLine(lineStr, str1, str2);

		if (strlen(str1) == 0 && strlen(str2) == 0)
			continue;
		else
		if (strcasecmp(str1, "sizeInKBytes") == 0)
			sscanf(str2, "%d", &sizeInKBytes);
	}
	fclose(inStream);
	return sizeInKBytes;
}

const File *Session::getNextTrack()
{
	File *nextTrack = NULL;

	DIR *dir = opendir(this->dir->getName());
	if (dir == NULL)
		return NULL;

	struct dirent *dirEntry;
	int n;
	while ((dirEntry = readdir(dir)) != NULL)
		if (sscanf(dirEntry->d_name, "track%d", &n) == 1)
			if (n == currentTrackNo + 1)
			{
				nextTrack = new File(
					this->dir, dirEntry->d_name);
				break;
			}

	closedir(dir);

	if (nextTrack == NULL)
		return NULL;

	if (nextTrack->exists())
	{
		if (currentTrack != NULL)
			delete currentTrack;
		currentTrack = nextTrack;
		currentTrackNo++;
		return currentTrack;
	}
	else
	{
		delete nextTrack;
		return NULL;
	}
}

int Session::isSessionDir(File &dir)
{
	File file(dir, CDWserverInfo);
	return file.exists();
}

void Session::makeDir(const char *fileName, int mode, int time)
{
	if (!isOpen)
		create();

	if (currentTrack == NULL
		|| !currentTrack->isDirectory())
		newTrack("dir");

	File dir(currentTrack, fileName);
	File dir2(dir, ".");
	dir2.mkdirs();

	struct utimbuf buf;
	buf.actime = buf.modtime = time;
	utime(dir.getName(), &buf);
}

void Session::newTrack(const char *trackType)
{
	if (mp3towav != NULL)
		closeMP3Track();

	const char *extension = "";
	if (strcasecmp(trackType, "image") == 0
		|| strcasecmp(trackType, "iso") == 0)
		extension = ".iso";
	else
	if (strcasecmp(trackType, "wav") == 0)
		extension = ".wav";
	else
	if (strcasecmp(trackType, "mp3") == 0)
	{
		extension = ".wav";
		mp3towav = new MP3toWAV();
	}

	tracksNum++;
	char nextTrackName[20];
	sprintf(nextTrackName, "track%03d%s", tracksNum, extension);
	currentTrack = new File(dir, nextTrackName);
	currentTrackNo = tracksNum;
}

void Session::open()
{
	if (!dir->exists())
		return;

	File file(dir, CDWserverInfo);
	FILE *inStream = fopen(file.getName(), "r");
	if (inStream == NULL)
		throw "internal error: can't read sessionInfo";

	char lineStr[1024];
	while (fgets(lineStr, sizeof lineStr, inStream) != NULL)
	{
		char str1[1024], str2[1024];
		Config::parseLine(lineStr, str1, str2);

		if (strlen(str1) == 0 && strlen(str2) == 0)
			continue;
		else
		if (strcasecmp(str1, "files") == 0)
			sscanf(str2, "%d", &filesNum);
		else
		if (strcasecmp(str1, "sizeInKBytes") == 0)
			sscanf(str2, "%d", &sizeInKBytes);
		else
		if (strcasecmp(str1, "tracks") == 0)
			sscanf(str2, "%d", &tracksNum);
	}
	fclose(inStream);

	isOpen = 1;
	log.put("opened previous session: "
		"files=%d, tracks=%d, sizeInKBytes=%d",
		filesNum, tracksNum, sizeInKBytes);
}

void Session::putFile(FILE *inStream,
	const char *fileName, int length, int mode, int time)
{
	if (!isOpen)
		create();

	if (strstr(fileName, "../") != NULL)
	{
		skipBytes(inStream, length);
		throw "filename contains ../";
	}

	if (currentTrack == NULL
		|| !currentTrack->isDirectory())
		newTrack("dir");

	File file(currentTrack, fileName);
	file.mkdirs();
	file.remove();

	FILE *outStream = fopen(file.getName(), "w");
	if (outStream == NULL)
	{
		skipBytes(inStream, length);
		log.put("Session::putFile: fopen");
		throw "cannot store file";
	}

	try
	{
		copy(inStream, outStream, length);
	}
	catch (const char *msg)
	{
		fclose(outStream);
		throw msg;
	}
	filesNum++;
	fclose(outStream);
	chmod(file.getName(), mode);
	struct utimbuf utimeBuf;
	utimeBuf.actime = utimeBuf.modtime = time;
	utime(file.getName(), &utimeBuf);
}

void Session::putTrack(FILE *inStream, int length)
{
	if (!isOpen)
		create();

	FILE *outStream = fopen(currentTrack->getName(), "a");
	if (outStream == NULL)
	{
		skipBytes(inStream, length);
		throw "cannot store track";
	}

	try
	{
		if (mp3towav != NULL)
		{
			mp3towav->setInStream(inStream);
			mp3towav->put(length);
			while (1)
			{
				const char *buf;
				int n = mp3towav->get(&buf);
				if (n == 0)
					break;
				try {
					copy(buf, outStream, n);
				}
				catch (const char *msg)
				{
					while (mp3towav->get(&buf) > 0);
					throw msg;
				}
			}
		}
		else
			copy(inStream, outStream, length);
	}
	catch (const char *msg)
	{
		fclose(outStream);
		throw msg;
	}
	fclose(outStream);
}

int Session::sizeToKBytes(int size)
{
	int KBytes = size / 1024;
	if (1024 * KBytes < size)
		KBytes++;
	return KBytes;
}

